/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>

#include "liberty.h"
#include "patterns.h"


static void compute_supplementary_dragon_data(void);
static void add_adjacent_dragons(int a, int b);
static void add_adjacent_dragon(int a, int b);
static void dragon_ring(int m, int n, int color, int *i, int *j);
static int compute_dragon_status(int i, int j);
static void dragon_eye(int m, int n, struct eye_data[MAX_BOARD][MAX_BOARD]);
static int compute_alternative_escape(int m, int n, int dragon_status_known);

/* This basic function finds all dragons and collects some basic information
 * about them in the dragon array.
 *
 * color is the player in turn to move. This does in no way affect the
 * information collected about the dragons, but it does affect what
 * information is passed on to the move generation code. If
 * color == EMPTY no information at all is passed on to the move generation.
 */

void 
make_dragons(int color)
{
  int m, n;
  int i, j;
  int d;

  /* We start with the dragon data copied from the worm data, then
   * modify it as the worms are amalgamated into larger dragons.
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      dragon[m][n].id             = -1;
      dragon[m][n].size           = worm[m][n].size;
      dragon[m][n].effective_size = worm[m][n].effective_size;
      dragon[m][n].color          = worm[m][n].color;
      dragon[m][n].origini        = worm[m][n].origini;
      dragon[m][n].originj        = worm[m][n].originj;
      dragon[m][n].lunchi         = -1;
      dragon[m][n].lunchj         = -1;
      dragon[m][n].mouse_attacki  = -1;
      dragon[m][n].mouse_attackj  = -1;
      dragon[m][n].mouse_defendi  = -1;
      dragon[m][n].mouse_defendj  = -1;
      dragon[m][n].escape_route   = worm[m][n].liberties4;
      dragon[m][n].heyes          = 0;
      dragon[m][n].genus          = 0;
      dragon[m][n].semeai         = 0;
				  
      half_eye[m][n].type         = 0;
      
      if (worm[m][n].origini == m && worm[m][n].originj == n)
	DEBUG(DEBUG_DRAGONS, 
	      "Initialising dragon from worm at %m, size %d, escape_route=%d\n", 
	      m, n, worm[m][n].size, worm[m][n].liberties4);
    }

  /* Amalgamate cavities. 
   *
   * Begin by finding the INESSENTIAL strings. These are defined as
   * surrounded strings which have no life potential unless part of
   * their surrounding chain can be captured. We give a conservative
   * definition of inessential: 
   *  - the genus must be zero 
   *  - there can no second order liberties
   *  - there can be no edge liberties
   *  - if it is removed from the board, the remaining cavity has
   *    border color the opposite color of the string 
   *  - it contains at most two edge vertices.
   *
   * An inessential string can be ignored for life and death purposes. It then
   * makes sense to amalgamate the surrounding cavities into a single cave
   * (empty dragon) with bordercolor the opposite of the inessential worm.
   *
   * For example, in the following situation:
   *
   *   OOOOO
   *   O.X.O
   *   OOOOO
   *
   * we find two graybordered cavities of size one. The X string is
   * inessential, so these two cavities are amalgamated into a single cave. 
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((p[m][n])
	  && (worm[m][n].origini == m)
	  && (worm[m][n].originj == n)
	  && (worm[m][n].genus == 0)
	  && (worm[m][n].liberties2 == 0)
	  && (worm[m][n].lunchi == -1)) 
      {
	int edge;
	int borigini=-1, boriginj=-1;

	int border_color = examine_cavity(m, n, &edge, NULL, NULL, NULL);
	if ((border_color != GRAY_BORDER) && (edge<3)) {
	  dragon_ring(m, n, EMPTY, &borigini, &boriginj);
	  worm[m][n].inessential=1;
	  propagate_worm(m, n);
	  dragon[borigini][boriginj].color=border_color;
	}
      }
    }

  make_domains(black_eye, white_eye);

  /* Find explicit connections patterns in database and amalgamate
   * involved dragons.
   */
  find_connections();
  
  /* Amalgamate dragons sharing an eyespace (not ko). At the same time
   * we decide to which dragon an eyespace belongs. Ko eyespaces
   * (typically false eyes but sometimes halfeyes) get assigned to an
   * arbitrary neighbor that is not the ko stone.
   */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {

      if ((black_eye[m][n].color == BLACK_BORDER) 
	  && (black_eye[m][n].origini == m)
	  && (black_eye[m][n].originj == n)) {
	if (!worm[m][n].ko
	   || (black_eye[m][n].esize > 1)) /* Only exclude living kos. */
	  dragon_eye(m, n, black_eye);
	else {
	  if (m > 0 && !worm[m-1][n].ko) {
	    black_eye[m][n].dragoni = dragon[m-1][n].origini;
	    black_eye[m][n].dragonj = dragon[m-1][n].originj;
	  } 
	  else if (m < board_size-1 && !worm[m+1][n].ko) {
	    black_eye[m][n].dragoni = dragon[m+1][n].origini;
	    black_eye[m][n].dragonj = dragon[m+1][n].originj;
	  } 
	  else if (n > 0 && !worm[m][n-1].ko) {
	    black_eye[m][n].dragoni = dragon[m][n-1].origini;
	    black_eye[m][n].dragonj = dragon[m][n-1].originj;
	  } 
	  else if (n < board_size-1 && !worm[m][n+1].ko) {
	    black_eye[m][n].dragoni = dragon[m][n+1].origini;
	    black_eye[m][n].dragonj = dragon[m][n+1].originj;
	  } 
	}
      }
	  
      if ((white_eye[m][n].color == WHITE_BORDER)
	  && (white_eye[m][n].origini == m)
	  && (white_eye[m][n].originj == n)) {
	if (!worm[m][n].ko
	    || (white_eye[m][n].esize > 1)) /* Only exclude living kos. */
	  dragon_eye(m, n, white_eye);
	else {
	  if (m > 0 && !worm[m-1][n].ko) {
	    white_eye[m][n].dragoni = dragon[m-1][n].origini;
	    white_eye[m][n].dragonj = dragon[m-1][n].originj;
	  } 
	  else if (m < board_size-1 && !worm[m+1][n].ko) {
	    white_eye[m][n].dragoni = dragon[m+1][n].origini;
	    white_eye[m][n].dragonj = dragon[m+1][n].originj;
	  } 
	  else if (n > 0 && !worm[m][n-1].ko) {
	    white_eye[m][n].dragoni = dragon[m][n-1].origini;
	    white_eye[m][n].dragonj = dragon[m][n-1].originj;
	  } 
	  else if (n < board_size-1 && !worm[m][n+1].ko) {
	    white_eye[m][n].dragoni = dragon[m][n+1].origini;
	    white_eye[m][n].dragonj = dragon[m][n+1].originj;
	  }
	}
      }
    }

  /* Find adjacent worms which can be easily captured: */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((worm[m][n].origini != m) 
	  || (worm[m][n].originj != n)
	  || p[m][n] == EMPTY
	  || (worm[m][n].lunchi == -1)) 
	continue;

      i=worm[m][n].lunchi;
      j=worm[m][n].lunchj;

      /* In contrast to worm lunches, a dragon lunch must also be
       * able to defend itself. 
       */
      if (worm[i][j].defend_code == 0)
	continue;

      /* Tell the move generation code about the lunch. */
      if (color != EMPTY)
	add_lunch(m, n, i, j);
	
      /* If several lunches are found, we pick the juiciest.
       * First maximize cutstone, then minimize liberties.
       */
      {
	int origini = dragon[m][n].origini;
	int originj = dragon[m][n].originj;

	if ((dragon[origini][originj].lunchi == -1)
	    || (worm[i][j].cutstone
		> worm[dragon[origini][originj].lunchi]
		[dragon[origini][originj].lunchj].cutstone)
	    || ((worm[i][j].cutstone 
		 == worm[dragon[origini][originj].lunchi]
		 [dragon[origini][originj].lunchj].cutstone) 
		&& (worm[i][j].liberties
		    < worm[dragon[m][n].lunchi][dragon[m][n].lunchj].liberties))) 
	  {
	    dragon[origini][originj].lunchi = worm[i][j].origini;
	    dragon[origini][originj].lunchj = worm[i][j].originj;
	    TRACE("at %m setting %m.lunch to %m (cutstone=%d)\n",
		  m, n, origini, originj,
		  worm[i][j].origini, worm[i][j].originj, worm[i][j].cutstone);
	  }
      }
    }

  /* Propagate lunch to rest of the dragon. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      dragon[i][j].lunchi
	= dragon[dragon[i][j].origini][dragon[i][j].originj].lunchi;
      dragon[i][j].lunchj
	= dragon[dragon[i][j].origini][dragon[i][j].originj].lunchj;
    }
  
  /* In case origins of dragons got moved, put the dragons of eyes aright. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (black_eye[i][j].dragoni != -1) {
	  int di=dragon[black_eye[i][j].dragoni]
                       [black_eye[i][j].dragonj].origini;
	  int dj=dragon[black_eye[i][j].dragoni]
	               [black_eye[i][j].dragonj].originj;
	  black_eye[i][j].dragoni=di;
	  black_eye[i][j].dragonj=dj;
      }

      if (white_eye[i][j].dragoni != -1) {
	  int di=dragon[white_eye[i][j].dragoni]
	               [white_eye[i][j].dragonj].origini;
	  int dj=dragon[white_eye[i][j].dragoni] 
	               [white_eye[i][j].dragonj].originj;
	  white_eye[i][j].dragoni=di;
	  white_eye[i][j].dragonj=dj;
      }
    }

  /* Find topological half eyes and false eyes by analyzing the
   * diagonal intersections, as described in the Texinfo
   * documentation (Eyes/Eye Topology).
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      int sum;
      int ai = -1;
      int aj = -1;
      int di = -1;
      int dj = -1;

      if (black_eye[m][n].color == BLACK_BORDER
	  && (!black_eye[m][n].marginal || life)
	  && (black_eye[m][n].neighbors <= 1)
	  && (black_eye[m][n].dragoni != -1)) {
	sum = topological_eye(m, n, BLACK, &ai, &aj, &di, &dj,
			      black_eye, white_eye, half_eye);
	if (sum >= 4) {
	  half_eye[m][n].type = FALSE_EYE;
	  if ((black_eye[m][n].esize == 1)
	      || legal(m, n, WHITE)
	      || p[m][n] == WHITE)
	    add_half_eye(m, n, black_eye, half_eye);
	}
	else if (sum == 3) {
	  half_eye[m][n].type = HALF_EYE;
#if 0
	  /* See comment in add_halfeye() */
	  half_eye[m][n].ai = ai;
	  half_eye[m][n].aj = aj;
	  half_eye[m][n].di = di;
	  half_eye[m][n].dj = dj;
	  add_half_eye(m, n, black_eye, half_eye);
#endif
	}
      }
      
      if (white_eye[m][n].color == WHITE_BORDER
	  && (!white_eye[m][n].marginal || life)
	  && (white_eye[m][n].neighbors <= 1)
	  && (white_eye[m][n].dragoni != -1)) {
	sum = topological_eye(m, n, WHITE, &ai, &aj, &di, &dj,
			      black_eye, white_eye, half_eye);
	if (sum >= 4) {
	  half_eye[m][n].type = FALSE_EYE;
	  if ((white_eye[m][n].esize == 1)
	      || legal(m, n, BLACK)
	      || p[m][n] == BLACK)
	    add_half_eye(m, n, white_eye, half_eye);
	}
	else if (sum == 3) {
	  half_eye[m][n].type = HALF_EYE;
#if 0
	  /* See comment in add_halfeye() */
	  half_eye[m][n].ai = ai;
	  half_eye[m][n].aj = aj;
	  half_eye[m][n].di = di;
	  half_eye[m][n].dj = dj;
	  add_half_eye(m, n, white_eye, half_eye);
#endif
	}
      }
    }

  /* Pattern based modification of the eye shapes computed by
   * make_domains and halfeye analysis.
   */
  modify_eye_spaces2();
  
  /* Compute the number of eyes, half eyes, etc. in an eye space. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((black_eye[i][j].color == BLACK_BORDER) 
	  && (black_eye[i][j].origini == i)
	  && (black_eye[i][j].originj == j)) 
      {
	int max, min, attacki, attackj, defendi, defendj;

	compute_eyes(i, j, &max, &min, &attacki, &attackj,
		     &defendi, &defendj, black_eye, half_eye, 1, color);
	DEBUG(DEBUG_EYES, "Black eyespace at %m: min=%d, max=%d\n",
	      i, j, min, max);
	black_eye[i][j].maxeye = max;
	black_eye[i][j].mineye = min;
	black_eye[i][j].attacki = attacki;
	black_eye[i][j].attackj = attackj;	  
	black_eye[i][j].defendi = defendi;
	black_eye[i][j].defendj = defendj;	  
	propagate_eye(i, j, black_eye);
      }

      if ((white_eye[i][j].color == WHITE_BORDER) 
	  && (white_eye[i][j].origini == i)
	  && (white_eye[i][j].originj == j)) 
      {
	int max, min, attacki, attackj, defendi, defendj;

	compute_eyes(i, j, &max, &min, &attacki, &attackj,
		     &defendi, &defendj, white_eye, half_eye, 1, color);
	DEBUG(DEBUG_EYES, "White eyespace at %m: min=%d, max=%d\n",
	      i, j, min, max);
	white_eye[i][j].maxeye = max;
	white_eye[i][j].mineye = min;
	white_eye[i][j].attacki = attacki;
	white_eye[i][j].attackj = attackj;	  
	white_eye[i][j].defendi = defendi;
	white_eye[i][j].defendj = defendj;	  
	propagate_eye(i, j, white_eye);
      }
    }

  /* Now we compute the genus. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((black_eye[i][j].color == BLACK_BORDER) 
	  && (black_eye[i][j].dragoni != -1) 
	  && (black_eye[i][j].origini == i)
	  && (black_eye[i][j].originj == j)) 
      {
	m = black_eye[i][j].dragoni;
	n = black_eye[i][j].dragonj;
	assert (p[m][n] == BLACK);
	TRACE("eye at %m found for dragon at %m--augmenting genus\n",
	      i, j, m, n);
	dragon[m][n].genus += (black_eye[i][j].mineye);
	dragon[m][n].heyes += (black_eye[i][j].maxeye-black_eye[i][j].mineye);
	if (black_eye[i][j].maxeye - black_eye[i][j].mineye > 0) {
	  dragon[m][n].heyei = black_eye[i][j].attacki;
	  dragon[m][n].heyej = black_eye[i][j].attackj;
	}
      }
      if ((white_eye[i][j].color == WHITE_BORDER) 
	  && (white_eye[i][j].dragoni != -1)
	  && (white_eye[i][j].origini == i)
	  && (white_eye[i][j].originj == j)) 
      {
	m = white_eye[i][j].dragoni;
	n = white_eye[i][j].dragonj;
	assert (p[m][n] == WHITE);
	TRACE("eye at %m found for dragon at %m--augmenting genus\n",
	      i, j, m, n);
	dragon[m][n].genus += (white_eye[i][j].mineye);
	dragon[m][n].heyes += (white_eye[i][j].maxeye-white_eye[i][j].mineye);
	if (white_eye[i][j].maxeye - white_eye[i][j].mineye > 0) {
	  dragon[m][n].heyei = white_eye[i][j].attacki;
	  dragon[m][n].heyej = white_eye[i][j].attackj;
	}
      }
    }

  /* Propagate genus to rest of the dragon. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      dragon[i][j].genus
	= dragon[dragon[i][j].origini][dragon[i][j].originj].genus;


  /* Recompute the escape route measure if alternative_escape option
   * is enabled.
   */
  if (alternative_escape) {
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) 
	if ((dragon[m][n].origini == m)
	    && (dragon[m][n].originj == n) && p[m][n] != EMPTY) {
	  dragon[m][n].escape_route = compute_alternative_escape(m, n, 0);
	}
  }

  /* Determine status: ALIVE, DEAD, CRITICAL or UNKNOWN */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if ((dragon[m][n].origini==m)
	  && (dragon[m][n].originj==n) && p[m][n])
	{
	  /* entirely ad-hoc : aim for -ve value to mean we need
	   * to worry about this dragon
	   * 2.5.38: Added effective_size to heuristics. /gf
	   */
	  dragon[m][n].vitality = 6 * dragon[m][n].genus 
	    + dragon[m][n].heyes + dragon[m][n].escape_route
	    + (dragon[m][n].effective_size - dragon[m][n].size)/6 - 10;
	  dragon[m][n].status = compute_dragon_status(m, n);
	  dragon[m][n].safety = dragon[m][n].status;
	  sgffile_dragon_status(m, n, dragon[m][n].status);
	}
    }


  compute_supplementary_dragon_data();
  
  /* Compute the surrounding moyo sizes. */
  for (d=0; d<number_of_dragons; d++) {
    dragon2[d].moyo = influence_get_moyo_size(dragon2[d].origini,
					      dragon2[d].originj, 1);
  }

  if (mouse) {
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) {
	int attacki = -1;
	int attackj = -1;
	int defendi = -1;
	int defendj = -1;
	  
	if (p[m][n] == EMPTY
	    || (dragon[m][n].origini != m)
	    || (dragon[m][n].originj != n))
	  continue;

	if (dragon[m][n].escape_route > 5 || DRAGON2(m, n).moyo > 10) {
	  dragon[m][n].mouse_status = UNCHECKED;
	  dragon[m][n].mouse_defendi = -1;
	  dragon[m][n].mouse_defendj = -1;
	  dragon[m][n].mouse_attacki = -1;
	  dragon[m][n].mouse_attackj = -1;
	}
	else {
	  if (mouse_attack(m, n, &attacki, &attackj)) {
	    dragon[m][n].mouse_attacki = attacki;
	    dragon[m][n].mouse_attackj = attackj;
	    if ((attacki != -1 )
		&& mouse_defend(m, n, &defendi, &defendj)) {
	      if (defendi != -1) {
		dragon[m][n].mouse_defendi = defendi;
		dragon[m][n].mouse_defendj = defendj;
		dragon[m][n].mouse_status = CRITICAL;
	      }
	    }
	    else {
	      dragon[m][n].mouse_status = DEAD; 
	      dragon[m][n].mouse_defendi = -1;
	      dragon[m][n].mouse_defendj = -1;
	    }
	  }
	  else {
	    dragon[m][n].mouse_status = ALIVE;
	    dragon[m][n].mouse_attacki = -1;
	    dragon[m][n].mouse_attackj = -1;
	  }
	}
      }
  }

  /* The dragon data is now correct at the origin of each dragon but
   * we need to copy it to every vertex.  
   */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      struct dragon_data *d = &(dragon[m][n]);
      dragon[m][n] = dragon[d->origini][d->originj];
    }
}

static void
compute_supplementary_dragon_data()
{
  int m, n;
  int i, j;
  int d;
  
  /* Give each dragon (caves excluded) an id number for indexing into
   * the dragon2 array. After this the DRAGON2 macro can be used.
   */
  number_of_dragons = 0;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if (p[m][n] == EMPTY)
	continue;
      i = dragon[m][n].origini;
      j = dragon[m][n].originj;
      if (dragon[i][j].id == -1)
	dragon[i][j].id = number_of_dragons++;
      dragon[m][n].id = dragon[i][j].id;
    }

  /* Now number_of_dragons contains the number of dragons and we can
   * allocate a dragon2 array of the appropriate size. First throw
   * away the old array.
   *
   * FIXME: As a future optimization we should only allocate a new
   *        array if the old one is too small.
   */
  if (dragon2 != NULL)
    free(dragon2);

  dragon2 = malloc(number_of_dragons * sizeof(*dragon2));
  assert(dragon2 != NULL);

  /* Find the origins of the dragons to establish the mapping back to
   * the board. After this the DRAGON macro can be used.
   */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if (p[m][n] != EMPTY
	  && dragon[m][n].origini == m
	  && dragon[m][n].originj == n) {
	DRAGON2(m, n).origini = m;
	DRAGON2(m, n).originj = n;
      }
    }

  /* Initialize the rest of the dragon2 data. */
  for (d=0; d<number_of_dragons; d++)
    dragon2[d].neighbors = 0;
  
  /* Examine which dragons are adjacent to each other. This is
   * complicated by the fact that adjacency may involve a certain
   * amount of empty space.
   *
   * The approach we use is to extend the dragons into their
   * surrounding influence areas until they collide. We also accept
   * one step extensions into neutral regions. After having done this
   * we can look for immediate adjacencies.
   */
  {
    int dragons[MAX_BOARD][MAX_BOARD];
    int distances[MAX_BOARD][MAX_BOARD];
    int dist;
    int k;
    int color;
    
    /* Initialize the arrays. */
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) {
	if (p[m][n] != EMPTY) {
	  dragons[m][n] = dragon[m][n].id;
	  distances[m][n] = 0;
	}
	else {
	  dragons[m][n] = -1;
	  distances[m][n] = -1;
	}
      }

    /* Expand from dist-1 to dist. Break out of the loop at the end if
     * we couldn't expand anything. Never expand more than five steps.
     */
    for (dist=1; dist<=5; dist++) {
      int found_one = 0;
      
      for (m=0; m<board_size; m++)
	for (n=0; n<board_size; n++) {
	  if (distances[m][n] != dist-1 || dragons[m][n] < 0)
	    continue;
	  color = DRAGON(dragons[m][n]).color;
	  for (k=0; k<4; k++) {
	    i = m + deltai[k];
	    j = n + deltaj[k];
	    if (!ON_BOARD(i, j))
	      continue;
	    /* Consider expansion from (m, n) to adjacent intersection
             * (i, j).
	     */
	    if (distances[i][j] >= 0 && distances[i][j] < dist)
	      continue; /* (i, j) already occupied. */
	    if (influence_area_color(m, n) == color
		&& influence_area_color(i, j) != OTHER_COLOR(color)) {
	      /* Expansion ok. Now see if someone else has tried to
               * expand here. In that case we indicate a collision by
               * setting the dragon number to -2.
	       */
	      if (distances[i][j] == dist) {
		if (dragons[i][j] != dragons[m][n])
		  dragons[i][j] = -2;
	      }
	      else {
		dragons[i][j] = dragons[m][n];
		distances[i][j] = dist;
		found_one = 1;
	      }
	    }
	  }
	}
      if (!found_one)
	break;
    }

    if (0) {
      for (m=0; m<board_size; m++) {
	for (n=0; n<board_size; n++) {
	  fprintf(stderr, "%3d", dragons[m][n]);
	}
	fprintf(stderr, "\n");
      }
      fprintf(stderr, "\n");
      
      for (m=0; m<board_size; m++) {
	for (n=0; n<board_size; n++) {
	  fprintf(stderr, "%3d", distances[m][n]);
	}
	fprintf(stderr, "\n");
      }
      fprintf(stderr, "\n");
    }

    /* Now go through dragons to find neighbors. It suffices to look
     * south and east for neighbors. In the case of a collision zone
     * where dragons==-2 we set all the neighbors of this intersection
     * as adjacent to each other.
     */
    for (m=0; m<board_size; m++)
      for (n=0; n<board_size; n++) {
	if (dragons[m][n] == -2) {
	  int neighbors = 0;
	  int adjacent[4];
	  for (k=0; k<4; k++) {
	    i = m + deltai[k];
	    j = n + deltaj[k];
	    if (ON_BOARD(i, j) && dragons[i][j] >= 0)
	      adjacent[neighbors++] = dragons[i][j];
	  }
	  for (i=0; i<neighbors; i++)
	    for (j=i+1; j<neighbors; j++)
	      add_adjacent_dragons(adjacent[i], adjacent[j]);
	}
	else if (dragons[m][n] >= 0) {
	  if (m < board_size-1) {
	    if (dragons[m+1][n] >= 0
		&& dragons[m+1][n] != dragons[m][n])
	      add_adjacent_dragons(dragons[m][n], dragons[m+1][n]);
	  }
	  if (n < board_size-1) {
	    if (dragons[m][n+1] >= 0
		&& dragons[m][n+1] != dragons[m][n])
	      add_adjacent_dragons(dragons[m][n], dragons[m][n+1]);
	  }
	}
      }

    if (0) {
      for (d=0; d<number_of_dragons; d++) {
	gprintf("dragon %d at %m:", d, dragon2[d].origini, dragon2[d].originj);
	for (i=0; i<dragon2[d].neighbors; i++)
	  gprintf(" %m(%d)", dragon2[dragon2[d].adjacent[i]].origini,
		  dragon2[dragon2[d].adjacent[i]].originj,
		  dragon2[d].adjacent[i]);
	gprintf("\n");
      }
    }
  }
}

/* Add the dragons with id a and b as adjacent to each other. */
static void
add_adjacent_dragons(int a, int b)
{
  assert(a >= 0 && a < number_of_dragons && b >= 0 && b < number_of_dragons);
  if (a == b)
    return;
  add_adjacent_dragon(a, b);
  add_adjacent_dragon(b, a);
}

/* Add the dragon with id b as adjacent to a. */
static void
add_adjacent_dragon(int a, int b)
{
  int i;
  assert(a >= 0 && a < number_of_dragons && b >= 0 && b < number_of_dragons);
  /* If the array of adjacent dragons already is full, ignore
   * additional neighbors.
   */
  if (dragon2[a].neighbors == MAX_NEIGHBOR_DRAGONS)
    return;
  
  for (i=0; i<dragon2[a].neighbors; i++)
    if (dragon2[a].adjacent[i] == b)
      return;

  dragon2[a].adjacent[dragon2[a].neighbors++] = b;
}

/* print status info on all dragons. (Can be invoked from gdb) 
 */
void 
show_dragons(void)
{
  static const char *cnames[] = 
    {"(empty)", "white dragon", "black dragon",
     "gray-bordered cave", "black-bordered cave", "white-bordered cave"};
  static const char *snames[] = 
    {"dead", "alive", "critical", "unknown", "unchecked"};

  int m, n;

  for (m=0;m<board_size;m++)
    for (n=0;n<board_size;n++) {
      struct worm_data *w = &(worm[m][n]);

      if (((w->origini)==m)&&((w->originj)==n)) {
	if (p[m][n]) {
	  gprintf("%m : (dragon %m) %s string of size %d (%f), genus %d: (%d,%d,%d,%d)",
		  m, n, dragon[m][n].origini, dragon[m][n].originj,
		  color_to_string(p[m][n]),
		  w->size,
		  w->effective_size,
		  w->genus,
		  w->liberties,
		  w->liberties2,
		  w->liberties3,
		  w->liberties4);
	  if (w->cutstone==1)
	    gprintf ("%o - is a potential cutting stone\n");
	  else if (w->cutstone==2)
	    gprintf("%o - is a cutting stone\n");
	  else gprintf("%o\n");

	  if (w->cutstone2 > 0)
	    gprintf("- cutstone2 = %d\n", w->cutstone2);
	  
	  if (w->attack_code != 0)
	    gprintf("- attackable at %m, attack code = %d\n",
		    w->attacki, w->attackj, w->attack_code);
	  if (w->defend_code != 0)
	    gprintf("- defendable at %m, defend code = %d\n",
		    w->defendi, w->defendj, w->defend_code);

	  if (w->inessential)
	    gprintf("- is inessential\n");
	  
	  if (w->ko == 1)
	    gprintf("- is a ko stone\n");
	}
	else 
	  gprintf("%m : cavity of size %d\n",m,n,w->size);
      }
    }

  gprintf("%o\n");
  for (m=0;m<board_size;m++)
    for (n=0;n<board_size;n++) {
      struct dragon_data *d = &(dragon[m][n]);
      struct dragon_data2 *d2 = &(dragon2[d->id]);
      int k;

      if (((d->origini)==m) && ((d->originj)==n)) {
	if (p[m][n]) {
	  gprintf("%m : %s dragon size %d (%f), genus %d, half eyes %d, escape factor %d, status %s, vitality %d%s, moyo size %d",
		  m,n,
		  p[m][n]==BLACK ? "B" : "W",
		  d->size,
		  d->effective_size,
		  d->genus,
		  d->heyes,
		  d->escape_route,
		  snames[d->status],
		  d->vitality,
		  d->safety == CRITICAL ? ", weak" : "",
		  d2->moyo);
	  if (mouse) {
	    gprintf(", mouse status %s\n", snames[d->mouse_status]);
	    if (d->mouse_status == CRITICAL) {
	      gprintf("... mouse attackable at %m\n",
		      d->mouse_attacki, d->mouse_attackj);
	      gprintf("... mouse defendable at %m\n",
		      d->mouse_defendi, d->mouse_defendj);
	    }
	  }
	  else
	    gprintf("\n");
	  gprintf("... neighbors");
	  for (k=0; k<d2->neighbors; k++) {
	    int d = d2->adjacent[k];
	    gprintf(" %m", dragon2[d].origini, dragon2[d].originj);
	  }
	  gprintf("\n");
	  if (d->lunchi != -1)
	    gprintf("... adjacent worm %m is lunch\n", d->lunchi, d->lunchj);
	}
	else {
	  gprintf("%m : cave of size %d and border color %s\n",
		  m,n,
		  d->size,
		  cnames[d->color]);
	  if (d->color == BLACK_BORDER || d->color == WHITE_BORDER) {
	    if (!worm[m][n].ko)
	      gprintf("... surrounded by dragon at %m\n",
		      d->borderi, d->borderj);
	    else
	      gprintf("... is a ko\n");
	  }
	}
      }
    }
}


  
/* The function dragon_ring(m, n, color, *di, *dj) amalgamates every 
 * dragon adjacent to the worm at (m, n) with the given color. The
 * amalgamated dragons then form a bigger dragon ringing the worm at
 * (m, n). 
 *
 * Amalgamation means that the individual dragons are to be combined
 * into the same dragon.  If color is not EMPTY, then p[m][n]==0 and
 * we have found an eye (possibly false). In this (*di, *dj) is set to
 * the origin of the surrounding dragon and dragon[*di][*dj].genus is
 * incremented.  
 */

static void
dragon_ring(int m, int n, int color, int *di, int *dj)
{
  int i, j;
  int dragoni=-1, dragonj=-1;

  DEBUG(DEBUG_DRAGONS, "amalgamate dragons around %m\n", m, n);
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((worm[i][j].origini==m) && (worm[i][j].originj==n)) {

	if ((i>0)&&(p[i-1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i-1][j].origini;
	    dragonj=dragon[i-1][j].originj;
	  } else if ((dragoni != dragon[i-1][j].origini) 
		     || (dragonj != dragon[i-1][j].originj))
	    join_dragons(i-1, j, dragoni, dragonj);
	}

	if ((i<board_size-1)&&(p[i+1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i+1][j].origini;
	    dragonj=dragon[i+1][j].originj;
	  } else if ((dragoni != dragon[i+1][j].origini) 
		     || (dragonj != dragon[i+1][j].originj))
	    join_dragons(i+1, j, dragoni, dragonj);
	}

	if ((j>0)&&(p[i][j-1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j-1].origini;
	    dragonj=dragon[i][j-1].originj;
	  } else if ((dragoni != dragon[i][j-1].origini) 
		     || (dragonj != dragon[i][j-1].originj))
	      join_dragons(i, j-1, dragoni, dragonj);
	}

	if ((j<board_size-1)&&(p[i][j+1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j+1].origini;
	    dragonj=dragon[i][j+1].originj;
	  } else if ((dragoni != dragon[i][j+1].origini) 
		     || (dragonj != dragon[i][j+1].originj))
	      join_dragons(i, j+1, dragoni, dragonj);
	}
      }
    }

  if (color) {
    ASSERT(p[m][n] == EMPTY, m, n); /* (m, n) is a cave */

    /* record the fact that we found an eye (could be false) */
    (dragon[dragoni][dragonj].genus)++;
  }
  *di=dragoni;   /* (m, n) is an inessential string. */
  *dj=dragonj;
}


/* 
 * dragon_eye(m, n, *di, *dj) is invoked with (m, n) the origin of an
 * eyespace. It unites all the worms adjacent to non-marginal points of
 * the eyespace into a single dragon with origin (*di, *dj). In addition
 * to marginal eye space points, amalgamation is inhibited for points
 * with the INHIBIT_CONNECTION type set.
 *
 * This is based on the older function dragon_ring. Unlike dragon_ring
 * this function does not attempt to keep track of the genus.
 */

static void
dragon_eye(int m, int n, struct eye_data eye[MAX_BOARD][MAX_BOARD])
{
  int i, j;
  int dragoni=-1, dragonj=-1;
  int color;

  /* don't amalgamate across ikken tobi */
  if ((eye[m][n].esize==3) && (eye[m][n].msize>1))
    return;

  DEBUG(DEBUG_DRAGONS, "amalgamate dragons around %m\n", m, n);
  if (eye[m][n].color==BLACK_BORDER)
    color=BLACK;
  else {
    assert(eye[m][n].color==WHITE_BORDER);
    color=WHITE;
  }

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((eye[i][j].origini==m)
	  && (eye[i][j].originj==n)
	  && !eye[i][j].marginal
	  && !(eye[i][j].type & INHIBIT_CONNECTION)) 
      {
	if ((i>0)&&(p[i-1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i-1][j].origini;
	    dragonj=dragon[i-1][j].originj;
	  } else if ((dragoni != dragon[i-1][j].origini) 
		     || (dragonj != dragon[i-1][j].originj))
	    join_dragons(i-1, j, dragoni, dragonj);
	}

	if ((i<board_size-1)&&(p[i+1][j]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i+1][j].origini;
	    dragonj=dragon[i+1][j].originj;
	  } else if ((dragoni != dragon[i+1][j].origini) 
		     || (dragonj != dragon[i+1][j].originj))
	    join_dragons(i+1, j, dragoni, dragonj);
	}

	if ((j>0)&&(p[i][j-1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j-1].origini;
	    dragonj=dragon[i][j-1].originj;
	  } else if ((dragoni != dragon[i][j-1].origini) 
		     || (dragonj != dragon[i][j-1].originj))
	      join_dragons(i, j-1, dragoni, dragonj);
	}

	if ((j<board_size-1)&&(p[i][j+1]==color)) {
	  if (dragoni == -1) {
	    dragoni=dragon[i][j+1].origini;
	    dragonj=dragon[i][j+1].originj;
	  } else if ((dragoni != dragon[i][j+1].origini) 
		     || (dragonj != dragon[i][j+1].originj))
	      join_dragons(i, j+1, dragoni, dragonj);
	}
      }
    }

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if ((eye[i][j].color == BLACK_BORDER 
	   || eye[i][j].color == WHITE_BORDER) 
	  && (eye[i][j].origini==m)
	  && (eye[i][j].originj==n)) 
      {
	eye[i][j].dragoni=dragoni;
	eye[i][j].dragonj=dragonj;
      }
    }
}


/* 
 * join_dragons amalgamates the dragon at (ai, aj) to the
 * dragon at (bi, bj).
 */

void 
join_dragons(int ai, int aj, int bi, int bj)
{
  int t, u;

  int i = dragon[ai][aj].origini;
  int j = dragon[ai][aj].originj;
  int m = dragon[bi][bj].origini;
  int n = dragon[bi][bj].originj;
  
  assert(worm[i][j].color==worm[m][n].color);

  DEBUG(DEBUG_DRAGONS,
	"joining dragon at %m with escape route %d to dragon at %m with %d\n", 
	i, j, dragon[i][j].escape_route, m, n, dragon[m][n].escape_route);

  dragon[m][n].size  = dragon[m][n].size + dragon[i][j].size;
  dragon[m][n].effective_size  = (dragon[m][n].effective_size
				  + dragon[i][j].effective_size);
  dragon[m][n].genus = dragon[m][n].genus + dragon[i][j].genus;
  if (dragon[i][j].escape_route > dragon[m][n].escape_route)
    dragon[m][n].escape_route = dragon[i][j].escape_route;

  for (t=0; t<board_size; t++)
    for (u=0; u<board_size; u++){
      if ((dragon[t][u].origini == i) && (dragon[t][u].originj == j)) {
	dragon[t][u].origini = m;
	dragon[t][u].originj = n;
      }
    }
}



/*
 * compute_dragon_status(i, j) tries to determine whether the dragon
 * at (i, j) is ALIVE, DEAD, or UNKNOWN. The algorithm is not perfect
 * and can give incorrect answers.
 *
 * The dragon is judged alive if its genus is >1. It is judged dead if
 * the genus is <2, it has no escape route, and no adjoining string can
 * be easily captured. Otherwise it is judged UNKNOWN.  */

static int 
compute_dragon_status(int i, int j)
{

  int true_genus = 2*dragon[i][j].genus + dragon[i][j].heyes;

  /* If it has two sure eyes, everything is just dandy. */
  if (true_genus > 3)
    return ALIVE;

  /* If the dragon consists of one worm, there is an attack, but 
   * no defense and there is less than one eye and one half eye,
   * the situation is hopeless. */
  if ((dragon[i][j].size==worm[i][j].size)
      && (worm[i][j].attack_code != 0) 
      && (worm[i][j].defend_code == 0)
      && (true_genus < 3))
    return DEAD;
  
  if ((dragon[i][j].lunchi != -1)
      && (true_genus < 3)
      && (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].defend_code != 0)
      && (dragon[i][j].escape_route <5))
    if ((true_genus == 2)
	|| (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].size >2))
      return CRITICAL;

  if ((dragon[i][j].lunchi != -1)
      && (true_genus >= 3))
    return ALIVE;

  if ((dragon[i][j].lunchi == -1)
      || (worm[dragon[i][j].lunchi][dragon[i][j].lunchj].cutstone <2)) 
  {
    if ((true_genus < 3) && (dragon[i][j].escape_route==0))
      return DEAD;

    if ((true_genus == 3) && (dragon[i][j].escape_route < 6))
      return CRITICAL;
  }

  return UNKNOWN;
}


/* Yet another dragon escape measure. This one is defined as follows.
 *   
 * Let a PATH be a sequence of adjacent intersections that do nowhere
 * touch or include an opponent stone or touch the border. It may
 * include friendly stones and those are allowed to touch opponent
 * stones or the border). Let a DISTANCE N INTERSECTION be an
 * intersection connected to a dragon by a path of length N, but by no
 * shorter path. The connection of the path to the dragon may either
 * be by direct adjacency or diagonally if both adjoining
 * intersections are empty.
 *
 * It is assumed that each intersection has an escape value, which
 * would typically depend on influence and (preliminary) dragon
 * status. We define the escape potential as the sum of the escape
 * values over the distance four intersections of the dragon.
 * 
 * Example of distance N intersections, 1 <= N <= 4:
 * 
 * . . . . . . . . .    . . . . . . . . .
 * . . . . . X . . O    . . . . . X . . O
 * . . X . . . . . O    . . X . 2 . 4 . O
 * X . . . . . . . .    X . . 1 1 2 3 4 .
 * X O . O . . . . O    X O 1 O 1 2 3 4 O
 * X O . O . . . . .    X O 1 O 1 . 4 . .
 * X O . . . X . O O    X O 1 . . X . . O
 * . . . X . . . . .    . 1 . X . . . . .
 * X . . . . X . . .    X . . . . X . . .
 * . . . . . . . . .    . . . . . . . . .
 * */

#define ENQUEUE(i, j) (queuei[queue_end] = (i),\
		       queuej[queue_end++] = (j),\
		       mx[i][j] = 1)

/* Compute the escape potential described above. The dragon is marked
 * in the goal array.
 */
int
dragon_escape(char goal[MAX_BOARD][MAX_BOARD], int color,
	      int escape_value[MAX_BOARD][MAX_BOARD])
{
  int i, j;
  int k;
  static int mx[MAX_BOARD][MAX_BOARD];
  static int mx_initialized = 0;
  int queuei[MAX_BOARD * MAX_BOARD];
  int queuej[MAX_BOARD * MAX_BOARD];
  int queue_start = 0;
  int queue_end = 0;
  int other = OTHER_COLOR(color);
  int distance;
  int escape_potential = 0;

  assert(color != EMPTY);
  
  if (!mx_initialized) {
    memset(mx, 0, sizeof(mx));
    mx_initialized = 1;
  }

  /* Enter the stones of the dragon in the queue. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if (goal[i][j])
	ENQUEUE(i, j);
  
  /* Find points at increasing distances from the dragon. At distance
   * four, sum the escape values at those points to get the escape
   * potential.
   */
  for (distance = 0; distance <= 4; distance++) {
    int save_queue_end = queue_end;
    while (queue_start < save_queue_end) {
      i = queuei[queue_start];
      j = queuej[queue_start];
      queue_start++;

      if (distance == 4)
	escape_potential += escape_value[i][j];
      else {
	if (i > 0
	    && !mx[i-1][j]
	    && (p[i-1][j] == color
		|| (p[i-1][j] == EMPTY
		    && i > 1 && p[i-2][j] != other
		    && j > 0 && p[i-1][j-1] != other
		    && j < board_size-1 && p[i-1][j+1] != other)))
	  ENQUEUE(i-1, j);

      	if (i < board_size-1
	    && !mx[i+1][j]
	    && (p[i+1][j] == color
		|| (p[i+1][j] == EMPTY
		    && i < board_size-2 && p[i+2][j] != other
		    && j > 0 && p[i+1][j-1] != other
		    && j < board_size-1 && p[i+1][j+1] != other)))
	  ENQUEUE(i+1, j);

	if (j > 0
	    && !mx[i][j-1]
	    && (p[i][j-1] == color
		|| (p[i][j-1] == EMPTY
		    && j > 1 && p[i][j-2] != other
		    && i > 0 && p[i-1][j-1] != other
		    && i < board_size-1 && p[i+1][j-1] != other)))
	  ENQUEUE(i, j-1);

	if (j < board_size-1
	    && !mx[i][j+1]
	    && (p[i][j+1] == color
		|| (p[i][j+1] == EMPTY
		    && j < board_size-2 && p[i][j+2] != other
		    && i > 0 && p[i-1][j+1] != other
		    && i < board_size-1 && p[i+1][j+1] != other)))
	  ENQUEUE(i, j+1);

	/* For distance one intersections, allow kosumi to move out. I.e.
	 *
	 * ??..
	 * X.*.
	 * ?O.?
	 * ??X?
	 *
	 */
	if (distance == 0) {
	  if (i > 1 && j > 1
	      && p[i][j-1] == EMPTY && p[i-1][j] == EMPTY
	      && (p[i-1][j-1] == color
		  || (p[i-1][j-1] == EMPTY
		      && p[i-2][j-1] != other && p[i-1][j-2] != other)))
	    ENQUEUE(i-1, j-1);

	  if (i > 1 && j < board_size-2
	      && p[i][j+1] == EMPTY && p[i-1][j] == EMPTY
	      && (p[i-1][j+1] == color
		  || (p[i-1][j+1] == EMPTY
		      && p[i-2][j+1] != other && p[i-1][j+2] != other)))
	    ENQUEUE(i-1, j+1);

	  if (i < board_size-2 && j < board_size-2
	      && p[i][j+1] == EMPTY && p[i+1][j] == EMPTY
	      && (p[i+1][j+1] == color
		  || (p[i+1][j+1] == EMPTY
		      && p[i+2][j+1] != other && p[i+1][j+2] != other)))
	    ENQUEUE(i+1, j+1);

	  if (i < board_size-2 && j > 1
	      && p[i][j-1] == EMPTY && p[i+1][j] == EMPTY
	      && (p[i+1][j-1] == color
		  || (p[i+1][j-1] == EMPTY
		      && p[i+2][j-1] != other && p[i+1][j-2] != other)))
	    ENQUEUE(i+1, j-1);
	}
      }
    }
  }

  /* Reset used mx cells. */
  for (k=0; k<queue_end; k++)
    mx[queuei[k]][queuej[k]] = 0;

  return escape_potential;
}

/* Wrapper to call the function above and compute the escape potential
 * for the dragon at (m, n). For the moment we give all intersections
 * escape value 1, but this has to be improved.
 */
static int
compute_alternative_escape(int m, int n, int dragon_status_known)
{
  int i, j;
  char goal[MAX_BOARD][MAX_BOARD];
  int escape_value[MAX_BOARD][MAX_BOARD];

  ASSERT(p[m][n] != EMPTY, m, n);
  
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      goal[i][j] = same_dragon(i, j, m, n);
    }

  compute_escape_influence(goal, p[m][n], escape_value, dragon_status_known);

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (dragon_status_known) {
	if (dragon[i][j].status == ALIVE)
	  escape_value[i][j] = 6;
	else if (dragon[i][j].status == UNKNOWN && dragon[i][j].vitality > 5)
	  escape_value[i][j] = 4;
      }
      else {
	if (p[i][j] == p[m][n]
	    && !goal[i][j]
	    && worm[i][j].attack_code == 0)
	  escape_value[i][j] = 2;
      }
    }

  return dragon_escape(goal, p[m][n], escape_value);
}

/* ================================================================ */
/*                       A few status functions                     */
/* ================================================================ */

/*
 * These functions are only here because then we don't need to expose
 * the dragon structure to the external program.
 */

int
dragon_status(int i, int j)
{
  return dragon[i][j].status;
}


/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
